;;;; structures.scm - SML-like module system - felix
;
; Copyright (c) 2000-2003, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Steinweg 1A
; 37130 Gleichen, OT Weissenborn
; Germany


(define structures:signature-table (make-hash-table eq?))

(define-macro (define-signature sig . items)
  (##sys#check-syntax 'define-signature sig 'symbol)
  `(eval-when (compile eval load)
     (hash-table-set!
      structures:signature-table ',sig
      (list ,@(map (match-lambda 
		     [((? symbol? s) pred) `(cons ',s ,pred)]
		     [(? symbol? s) `',s]
		     [x (error "invalid signature item" x)] )
		   items) ) ) ) )

(define-macro (structure exports . defs)
  (let ([vars '()]
	[ops '()] )
    (let rec ([defs defs])
      (if (null? defs)
	  `(let ,(map (lambda (v) `(,v (##sys#void))) vars)
	     ,@(reverse ops)
	     (structures:make-structure 
	      (list ,@(map (lambda (v) `(cons ',v ,v)) vars))
	      ',exports) )
	  (match (car defs)
	    [('define ((? symbol? s) . llist) . body)
	     (set! vars (cons s vars))
	     (set! ops (cons `(set! ,s (lambda ,llist ,@body)) ops))
	     (rec (cdr defs)) ]
	    [('define (? symbol? s) x)
	     (set! vars (cons s vars))
	     (set! ops (cons `(set! ,s ,x) ops))
	     (rec (cdr defs)) ]
	    [('define-values ((? symbol? s) ...) x)
	     (set! vars (append s vars))
	     (set! ops 
	       (cons
		(let ([tmps (map gensym s)])
		  `(##sys#call-with-values 
		    (lambda () ,x)
		    (lambda ,tmps ,@(map (lambda (v t) `(set! ,v ,t)) s tmps)) ) )
		ops) )
	     (rec (cdr defs)) ]
	    [def 
	     (set! ops (cons def ops))
	     (rec (cdr defs)) ] ) ) ) ) )

(define-macro (in-structure struct imports . body)
  (let loop ([imps imports] [impvars '()])
    (if (null? imps)
	(let ([svar (gensym)]
	      [impvars (delete-duplicates impvars eq?)] ) ; should we check for errors?
	  `(let ([,svar ,struct])
	     (let ,(map (lambda (v) `(,v (structures:structure-ref ,svar ',v))) impvars)
	       ,@body) ) )
	(match (car imps)
	  [('import (? symbol? s) ...)
	   (loop (cdr imps) (append s impvars)) ]
	  [(? symbol? sig)
	   (loop (cdr imps) 
		 (append (or (hash-table-ref/default structures:signature-table sig #f)
			     (error "undefined signature" sig) )
			 impvars) ) ]
	  [x (error "invalid import syntax" x)] ) ) ) )

(define-macro (functor srefs exports . defs)
  `(let ,(map (match-lambda
		[((? symbol? var) struct . imports) 
		 `(,var (let ([tmp ,struct])
			  (structures:restrict-structure
			   tmp
			   '(,@(map (match-lambda
				      [('import (? symbol? s) ...) `(export ,@s)]
				      [(? symbol? sig) `,sig]
				      [x (error "invalid import syntax" x)] )
				    imports) ) ) ) ) ]
		[x (error "invalid functor argument syntax" x)] )
	      srefs)
     ,(let fold ([srefs srefs])
	(if (null? srefs)
	    `(structure ,exports ,@defs)
	    `(in-structure ,(caar srefs) ,(cddar srefs)
	       ,(fold (cdr srefs)) ) ) ) ) )

(define-macro (define-functor head exports . defs)
  (match head
    [((? symbol? name) . srefs)
     (let ([sdefs
	    (map (match-lambda
		   [(and x ((? symbol? var) . imports)) x]
		   [(? symbol? var) (list var)]
		   [x (error "invalid functor argument syntax" x)] )
		 srefs) ] )
       `(define (,name ,@(map car sdefs))
	  (functor ,(map (lambda (s) (cons (car s) s)) sdefs) 
	      ,exports
	    ,@defs) ) ) ]
    [_ (error "invalid syntax in functor definition" x)] ) )
